/* Daemonise.c
 *  Simple program to fork() into the background and redirect all input and
 *  output to /dev/null, before replacing the image of the child process with
 *  that of the command passed.  This will return the process ID as a string
 *  to stdout if successful.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>


static void usage(const char *imageName)
{
    printf("Usage: %s [-u user] command [args]\n", imageName);
    printf("\n");
    printf(" -u user  The user to switch to prior to running the command.\n");
    printf(" command  A command to execute.\n");
    printf(" args     Optional list of parameters to pass to the command.\n");
    printf("\n");
    printf("Description\n");
    printf(" Creates a detached child process with stdin, stdout and stderr \n");
    printf(" redirected to /dev/null, and prints the PID of the child process\n");
    printf(" to stdout.  The child process will be the command specified, which\n");
    printf(" should have it's full path listed.\n");
    printf(" If '-u' and a user are specified, the program will check if the\n");
    printf(" current user matches the passed name, and will attempt to switch\n");
    printf(" to that user if possible.  If there are insufficient permissions\n");
    printf(" to switch to the specified user, or that user cannot be resolved\n");
    printf(" by the passwd file, an error message will be printed and the program\n");
    printf(" will return 1 to indicate failure.  On success, 0 is returned.\n");
    printf("\n");
}


int main(int argc, char *argv[], char *envp[])
{

    if(argc < 2 || (strcmp(argv[1], "-u") == 0 && argc < 4))
    {
        usage(argv[0]);
    }
    else
    {
        pid_t child;
        int   cmdArg;

        /* Check if the user should be verified and optionally switched */
        if(strcmp(argv[1], "-u") == 0)
        {
            struct passwd *pinfo;
            char           err[80];

            /* Convert username to UID */
            pinfo = getpwnam(argv[2]);
            if(pinfo == NULL)
            {
                snprintf(err, sizeof(err), "Failed to retrieve passwd data for user '%s'", argv[2]);
                perror(err);
                return 1;
            }

            /* Check if the user is different to that desired */
            if(getuid() != pinfo->pw_uid)
            {
                if(setuid(pinfo->pw_uid) != 0)
                {
                    snprintf(err, sizeof(err), "Failed to setuid() to '%s'", argv[2]);
                    perror(err);
                    return 1;
                }
            }

            cmdArg = 3;

        }
        else
        {
            cmdArg = 1;
        }

        /* Fork into background */
        child = fork();
        if(child == 0)
        {
            int stderrfd, fd;

            /* Become group leader */
            if(setsid() < 0)
            {
                perror("setsid failed");
            }

            /* Save stderr incase it is needed later.
             *  Used close-on-exec to allow the process image that takes over
             *  at execve() to be fully detached, but to allow the handle to
             *  still exist for error reporting if execve() fails.
             */
            stderrfd = dup(2);
            if(fcntl(stderrfd, F_SETFD, FD_CLOEXEC) == -1)
            {
                perror("failed to set file descriptor to close-on-exec");
            }

            /* Redirect streams */
            fd = open("/dev/null", O_RDWR);
            if(fd < 0)
            {
                perror("open failed for /dev/null");
            }
            else
            {
                dup2(fd, 0);
                dup2(fd, 1);
                dup2(fd, 2);
            }

            /* Call the program */
            if(execve(argv[cmdArg], &argv[cmdArg], envp) == -1)
            {
                char err[80];

                /* Restore stderr to display the message */
                dup2(stderrfd, 2);

                snprintf(err, sizeof(err), "execv() failed to run '%s'", argv[cmdArg]);
                perror(err);

                return 1;
            }

            /* Unreachable; execve never returns */

        }
        else
        {
            printf("%d\n", child);
        }
    }

    return 0;
}


/* END OF FILE */

